from typing import Any, Callable, Dict

import torch
from torch import Tensor, nn
from torch.utils.data import DataLoader, Dataset
from tqdm.auto import tqdm

from oml.ddp.patching import patch_dataloader_to_ddp
from oml.ddp.utils import get_world_size_safe, is_ddp, sync_dicts_ddp
from oml.utils.misc_torch import (
    drop_duplicates_by_ids,
    get_device,
    temporary_setting_model_mode,
)


@torch.no_grad()
def _inference(
    model: nn.Module,
    apply_model: Callable[[nn.Module, Dict[str, Any]], Tensor],
    dataset: Dataset,  # type: ignore
    num_workers: int,
    batch_size: int,
    verbose: bool,
    use_fp16: bool,
    accumulate_on_cpu: bool = True,
) -> Tensor:
    assert hasattr(dataset, "index_key"), "We expect that your dataset returns samples ids in __getitem__ method"

    loader = DataLoader(dataset=dataset, batch_size=batch_size, num_workers=num_workers, shuffle=False)

    if is_ddp():
        loader = patch_dataloader_to_ddp(loader)

    if verbose:
        loader = tqdm(loader, desc=str(get_device(model)))

    outputs_list = []
    ids = []

    with torch.autocast(device_type="cuda", dtype=torch.float16 if use_fp16 else torch.float32):
        with temporary_setting_model_mode(model, set_train=False):
            for batch in loader:
                out = apply_model(model, batch)
                if accumulate_on_cpu:
                    out = out.cpu()
                outputs_list.append(out)
                ids.extend(batch[dataset.index_key].long().tolist())

    outputs = torch.cat(outputs_list).detach()

    data_to_sync = {"outputs": outputs, "ids": ids}
    data_synced = sync_dicts_ddp(data_to_sync, world_size=get_world_size_safe())
    outputs, ids = data_synced["outputs"], data_synced["ids"]

    ids, outputs = drop_duplicates_by_ids(ids=ids, data=outputs, sort=True)

    assert len(outputs) == len(dataset), "Data was not collected correctly after DDP sync."
    assert list(range(len(dataset))) == ids, "Data was not collected correctly after DDP sync."

    return outputs


__all__ = ["_inference"]



from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from torch import Tensor

from oml.const import INDEX_KEY, PAIR_1ST_KEY, PAIR_2ND_KEY
from oml.datasets.list_dataset import ListDataset, TBBoxes
from oml.interfaces.datasets import IPairsDataset
from oml.transforms.images.torchvision import get_normalisation_torch
from oml.transforms.images.utils import TTransforms
from oml.utils.images.images import TImReader, imread_pillow


class EmbeddingPairsDataset(IPairsDataset):
    """
    Dataset to iterate over pairs of embeddings.

    """

    def __init__(
        self,
        embeddings1: Tensor,
        embeddings2: Tensor,
        pair_1st_key: str = PAIR_1ST_KEY,
        pair_2nd_key: str = PAIR_2ND_KEY,
        index_key: str = INDEX_KEY,
    ):
        """

        Args:
            embeddings1: The first input embeddings
            embeddings2: The second input embeddings
            pair_1st_key: Key to put ``embeddings1`` into the batches
            pair_2nd_key: Key to put ``embeddings2`` into the batches
            index_key: Key to put samples' ids into the batches

        """
        assert embeddings1.shape == embeddings2.shape
        assert embeddings1.ndim >= 2

        self.pair_1st_key = pair_1st_key
        self.pair_2nd_key = pair_2nd_key
        self.index_key = index_key

        self.embeddings1 = embeddings1
        self.embeddings2 = embeddings2

    def __getitem__(self, idx: int) -> Dict[str, Tensor]:
        return {self.pair_1st_key: self.embeddings1[idx], self.pair_2nd_key: self.embeddings2[idx], self.index_key: idx}

    def __len__(self) -> int:
        return len(self.embeddings1)


class ImagePairsDataset(IPairsDataset):
    """
    Dataset to iterate over pairs of images.

    """

    def __init__(
        self,
        paths1: List[Path],
        paths2: List[Path],
        bboxes1: Optional[TBBoxes] = None,
        bboxes2: Optional[TBBoxes] = None,
        transform: Optional[TTransforms] = None,
        f_imread: TImReader = imread_pillow,
        pair_1st_key: str = PAIR_1ST_KEY,
        pair_2nd_key: str = PAIR_2ND_KEY,
        index_key: str = INDEX_KEY,
        cache_size: Optional[int] = 0,
    ):
        """
        Args:
            paths1: Paths to the 1st input images
            paths2: Paths to the 2nd input images
            bboxes1: Should be either ``None`` or a sequence of bboxes.
                If an image has ``N`` boxes, duplicate its
                path ``N`` times and provide bounding box for each of them.
                If you want to get an embedding for the whole image, set bbox to ``None`` for
                this particular image path. The format is ``x1, y1, x2, y2``.
            bboxes2: The same as ``bboxes2``, but for the second inputs.
            transform: Augmentations for the images, set ``None`` to perform only normalisation and casting to tensor
            f_imread: Function to read the images
            pair_1st_key: Key to put the 1st images into the batches
            pair_2nd_key: Key to put the 2nd images into the batches
            index_key: Key to put samples' ids into the batches
            cache_size: Size of the dataset's cache

        """
        assert len(paths1) == len(paths2)

        if transform is None:
            transform = get_normalisation_torch()

        cache_size = cache_size // 2 if cache_size else None
        dataset_args = {"transform": transform, "f_imread": f_imread, "cache_size": cache_size}
        self.dataset1 = ListDataset(paths1, bboxes=bboxes1, **dataset_args)
        self.dataset2 = ListDataset(paths2, bboxes=bboxes2, **dataset_args)

        self.pair_1st_key = pair_1st_key
        self.pair_2nd_key = pair_2nd_key
        self.index_key = index_key

    def __getitem__(self, idx: int) -> Dict[str, Union[int, Dict[str, Any]]]:
        return {self.pair_1st_key: self.dataset1[idx], self.pair_2nd_key: self.dataset2[idx], self.index_key: idx}

    def __len__(self) -> int:
        return len(self.dataset1)


__all__ = ["EmbeddingPairsDataset", "ImagePairsDataset"]



from typing import Any, Dict, List, Sequence, Union

import numpy as np
import torch
from torch.distributed import get_world_size

from oml.ddp.utils import is_ddp, sync_dicts_ddp

TStorage = Dict[str, Union[torch.Tensor, np.ndarray, List[Any]]]


class Accumulator:
    def __init__(self, keys_to_accumulate: Sequence[str]):
        """
        Class for accumulating values of different types, for instance,
        torch.Tensor and numpy.array.

        Args:
            keys_to_accumulate: List or tuple of keys to be collected.
                                 We will take values via these keys calling ``self.update_data()``.
        """
        assert len(keys_to_accumulate) == len(set(keys_to_accumulate)), "All the keys have to be unique!"

        self.keys_to_accumulate = keys_to_accumulate
        self.num_samples = None

        self._collected_samples = 0
        self._storage: TStorage = dict()

    def refresh(self, num_samples: int) -> None:
        """
        This method refreshes the state.

        Args:
            num_samples:  The total number of elements you are going to collect (for memory allocation)
        """
        assert isinstance(num_samples, int) and num_samples > 0
        self.num_samples = num_samples  # type: ignore
        self._collected_samples = 0
        self._storage = {}

    def _allocate_memory_if_need(self, key: str, batch_value: Any) -> None:
        if self.num_samples is None:
            raise ValueError(
                f"The parameter for memory allocation has not been set up."
                f"Are you sure you've called {self.refresh.__name__}?"
            )

        if key not in self._storage:
            if isinstance(batch_value, torch.Tensor):
                self._storage[key] = torch.empty(
                    (self.num_samples, *batch_value.shape[1:]),
                    dtype=batch_value.dtype,
                    device="cpu",
                    requires_grad=False,
                )
            elif isinstance(batch_value, np.ndarray):
                self._storage[key] = np.empty((self.num_samples, *batch_value.shape[1:]), dtype=batch_value.dtype)
            elif isinstance(batch_value, (list, tuple)):
                self._storage[key] = []
            else:
                raise TypeError(f"Type '{type(batch_value)}' is not available for accumulating")

    def _put_in_storage(self, key: str, batch_value: Any) -> None:
        bs = len(batch_value)

        if isinstance(batch_value, torch.Tensor):
            bv = batch_value.detach()
            self._storage[key][self._collected_samples : self._collected_samples + bs, ...] = bv  # type: ignore
        elif isinstance(batch_value, np.ndarray):
            bv = batch_value
            self._storage[key][self._collected_samples : self._collected_samples + bs, ...] = bv  # type: ignore
        elif isinstance(batch_value, (list, tuple)):
            self._storage[key].extend(list(batch_value))  # type: ignore
        else:
            raise TypeError(f"Type '{type(batch_value)}' is not available for accumulating")

    def update_data(self, data_dict: Dict[str, Any]) -> None:
        """
        Args:
            data_dict: We will accumulate data getting values via ``self.keys_to_accumulate``.

        """
        bs_values = [len(data_dict[k]) for k in self.keys_to_accumulate]
        bs = bs_values[0]
        assert all(bs == bs_value for bs_value in bs_values), f"Lengths of data are not equal, lengths: {bs_values}"

        for k in self.keys_to_accumulate:
            v = data_dict[k]
            self._allocate_memory_if_need(k, v)
            self._put_in_storage(k, v)
        self._collected_samples += bs

    @property
    def storage(self) -> TStorage:
        return self._storage

    @property
    def collected_samples(self) -> int:
        return self._collected_samples

    def is_storage_full(self) -> bool:
        return self.num_samples == self.collected_samples

    def sync(self) -> "Accumulator":
        # TODO: add option to broadcast instead of sync to avoid duplicating data
        if not self.is_storage_full():
            raise ValueError("Only full storages could be synced")

        if is_ddp():
            world_size = get_world_size()
            if world_size == 1:
                return self
            else:
                params = {"num_samples": [self.num_samples], "keys_to_accumulate": self.keys_to_accumulate}

                gathered_params = sync_dicts_ddp(params, world_size=world_size, device="cpu")
                gathered_storage = sync_dicts_ddp(self._storage, world_size=world_size, device="cpu")

                assert set(gathered_params["keys_to_accumulate"]) == set(
                    self.keys_to_accumulate
                ), "Keys of accumulators should be the same on each device"

                synced_accum = Accumulator(list(set(gathered_params["keys_to_accumulate"])))
                synced_accum.refresh(sum(gathered_params["num_samples"]))
                synced_accum.update_data(gathered_storage)

                return synced_accum
        else:
            return self


__all__ = ["TStorage", "Accumulator"]



import logging
import warnings
from functools import lru_cache
from itertools import chain
from pathlib import Path
from random import sample
from typing import Any, Dict, List, Optional, Tuple, Union

import albumentations as albu
import numpy as np
import torch
from torch.utils.data import Dataset

from oml.const import INDEX_KEY, INPUT_TENSORS_KEY
from oml.transforms.images.albumentations import get_normalisation_albu
from oml.utils.images.images import TImReader, imread_cv2

TPath = Union[Path, str]
TTriplet = Tuple[TPath, TPath, TPath]
TItem = Dict[str, Any]


class TriDataset(Dataset):
    def __init__(
        self,
        triplets: List[TTriplet],
        im_root: Path,
        transforms: albu.Compose,
        expand_ratio: float,
        f_imread: TImReader = imread_cv2,
        cache_size: Optional[int] = 0,
        index_key: str = INDEX_KEY,
    ):
        """

        Args:
            triplets: List of triplets
            im_root: Images directory
            transforms: Image transforms
            expand_ratio: Set expand_ratio > 0 to generate additional triplets.
                          We keep positive pairs, but generale negative ones randomly.
                          After this procedure you dataset's length will increased (1 + expand_ratio) times
            f_imread: Function to read image from disk
            cache_size: Size of the dataset's cache
            index_key: Key to put samples' ids into the batches

        """
        assert expand_ratio >= 0

        warnings.warn("For now we do not keep this dataset up to date. Use it carefully.")

        self.triplets = triplets

        self.all_ims = set(chain(*triplets))

        self.im_root = im_root
        self.expand_ratio = expand_ratio
        self.f_imread = f_imread

        self.transforms = transforms or get_normalisation_albu()
        assert isinstance(transforms, albu.Compose) or (transforms is None)

        self.read_bytes_image = (
            lru_cache(maxsize=cache_size)(self._read_bytes_image) if cache_size else self._read_bytes_image
        )

        self.index_key = index_key

        logging.info(f"Dataset contains {len(self.triplets)} triplets.")

    @staticmethod
    def _read_bytes_image(path: Union[Path, str]) -> bytes:
        with open(str(path), "rb") as fin:
            return fin.read()

    def get_image(self, path: Union[Path, str]) -> np.ndarray:
        image_bytes = self.read_bytes_image(path)  # type: ignore
        image = self.f_imread(image_bytes)
        return image

    def __len__(self) -> int:
        return int((1 + self.expand_ratio) * len(self.triplets))

    def __getitem__(self, idx: int) -> TItem:  # type: ignore
        if idx < len(self.triplets):
            triplet = self.triplets[idx]
        else:
            # here we randomly create negative pair for the picked positive one
            a, p, _ = sample(self.triplets, k=1)[0]
            n = sample(self.all_ims - {a, p}, k=1)[0]
            triplet = a, p, n

        assert len(triplet) == 3

        images = tuple(map(lambda x: self.get_image(self.im_root / x), triplet))

        tensors = tuple(map(lambda x: self.transforms(image=x)["image"], images))

        tri_ids = (f"{idx}_a", f"{idx}_p", f"{idx}_n")
        return {INPUT_TENSORS_KEY: tensors, "tri_ids": tri_ids, "images": images, self.index_key: idx}


def tri_collate(items: List[TItem]) -> Dict[str, Any]:
    batch = dict()

    for key in (INPUT_TENSORS_KEY,):
        batch[key] = torch.stack(list(chain(*[item[key] for item in items])))

    for key in ("tri_ids", "images"):
        if key in items[0].keys():
            batch[key] = list(chain(*[item[key] for item in items]))

    return batch


__all__ = ["TPath", "TTriplet", "TItem", "TriDataset", "tri_collate"]



from pathlib import Path
from typing import Any, Dict, List, Optional

from torch import Tensor

from oml.datasets.pairs import EmbeddingPairsDataset, ImagePairsDataset
from oml.inference.abstract import _inference
from oml.interfaces.models import IPairwiseModel
from oml.transforms.images.utils import TTransforms
from oml.utils.images.images import TImReader
from oml.utils.misc_torch import get_device


def pairwise_inference_on_images(
    model: IPairwiseModel,
    paths1: List[Path],
    paths2: List[Path],
    transform: TTransforms,
    num_workers: int,
    batch_size: int,
    verbose: bool = True,
    f_imread: Optional[TImReader] = None,
    use_fp16: bool = False,
    accumulate_on_cpu: bool = True,
) -> Tensor:
    device = get_device(model)

    dataset = ImagePairsDataset(
        paths1=paths1,
        paths2=paths2,
        transform=transform,
        f_imread=f_imread,
        cache_size=0,
    )

    def _apply(
        model_: IPairwiseModel,
        batch_: Dict[str, Any],
    ) -> Tensor:
        pair1 = batch_[dataset.pair_1st_key][dataset.dataset1.input_tensors_key].to(device)
        pair2 = batch_[dataset.pair_2nd_key][dataset.dataset2.input_tensors_key].to(device)
        return model_.predict(pair1, pair2)

    output = _inference(
        model=model,
        apply_model=_apply,
        dataset=dataset,
        num_workers=num_workers,
        batch_size=batch_size,
        verbose=verbose,
        use_fp16=use_fp16,
        accumulate_on_cpu=accumulate_on_cpu,
    )

    return output


def pairwise_inference_on_embeddings(
    model: IPairwiseModel,
    embeddings1: Tensor,
    embeddings2: Tensor,
    num_workers: int,
    batch_size: int,
    verbose: bool = False,
    use_fp16: bool = False,
    accumulate_on_cpu: bool = True,
) -> Tensor:
    device = get_device(model)

    dataset = EmbeddingPairsDataset(embeddings1=embeddings1, embeddings2=embeddings2)

    def _apply(
        model_: IPairwiseModel,
        batch_: Dict[str, Any],
    ) -> Tensor:
        pair1 = batch_[dataset.pair_1st_key].to(device)
        pair2 = batch_[dataset.pair_2nd_key].to(device)
        return model_.predict(pair1, pair2)

    output = _inference(
        model=model,
        apply_model=_apply,
        dataset=dataset,
        num_workers=num_workers,
        batch_size=batch_size,
        verbose=verbose,
        use_fp16=use_fp16,
        accumulate_on_cpu=accumulate_on_cpu,
    )

    return output


__all__ = [
    "pairwise_inference_on_images",
    "pairwise_inference_on_embeddings",
]



from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union

import albumentations as albu
import numpy as np
import pandas as pd
import torchvision
from torch.utils.data import Dataset

from oml.const import (
    CATEGORIES_COLUMN,
    CATEGORIES_KEY,
    INDEX_KEY,
    INPUT_TENSORS_KEY,
    IS_GALLERY_COLUMN,
    IS_GALLERY_KEY,
    IS_QUERY_COLUMN,
    IS_QUERY_KEY,
    LABELS_COLUMN,
    LABELS_KEY,
    PATHS_COLUMN,
    PATHS_KEY,
    SEQUENCE_COLUMN,
    SEQUENCE_KEY,
    SPLIT_COLUMN,
    X1_COLUMN,
    X1_KEY,
    X2_COLUMN,
    X2_KEY,
    Y1_COLUMN,
    Y1_KEY,
    Y2_COLUMN,
    Y2_KEY,
)
from oml.interfaces.datasets import IDatasetQueryGallery, IDatasetWithLabels
from oml.registry.transforms import get_transforms
from oml.transforms.images.utils import TTransforms, get_im_reader_for_transforms
from oml.utils.dataframe_format import check_retrieval_dataframe_format
from oml.utils.images.images import TImReader


class BaseDataset(Dataset):
    """
    Base class for the retrieval datasets.

    """

    def __init__(
        self,
        df: pd.DataFrame,
        extra_data: Optional[Dict[str, Any]] = None,
        transform: Optional[TTransforms] = None,
        dataset_root: Optional[Union[str, Path]] = None,
        f_imread: Optional[TImReader] = None,
        cache_size: Optional[int] = 0,
        input_tensors_key: str = INPUT_TENSORS_KEY,
        labels_key: str = LABELS_KEY,
        paths_key: str = PATHS_KEY,
        categories_key: Optional[str] = CATEGORIES_KEY,
        sequence_key: Optional[str] = SEQUENCE_KEY,
        x1_key: str = X1_KEY,
        x2_key: str = X2_KEY,
        y1_key: str = Y1_KEY,
        y2_key: str = Y2_KEY,
        index_key: str = INDEX_KEY,
    ):
        """

        Args:
            df: Table with the following obligatory columns:

                  ``LABELS_COLUMN``, ``PATHS_COLUMN``

                  and the optional ones:

                  ``X1_COLUMN``, ``X2_COLUMN``, ``Y1_COLUMN``, ``Y2_COLUMN``, ``CATEGORIES_COLUMN``

            extra_data: Dictionary with additional information which we want to put into batches. We assume that
                the length of each record in this structure is the same as dataset's size.
            transform: Augmentations for the images, set ``None`` to perform only normalisation and casting to tensor
            dataset_root: Path to the images' dir, set ``None`` if you provided the absolute paths in your dataframe
            f_imread: Function to read the images, pass ``None`` so we pick it autmatically based on provided transforms
            cache_size: Size of the dataset's cache
            input_tensors_key: Key to put tensors into the batches
            labels_key: Key to put labels into the batches
            paths_key: Key put paths into the batches
            categories_key: Key to put categories into the batches
            sequence_key: Key to put sequence ids into the batches
            x1_key: Key to put ``x1`` into the batches
            x2_key: Key to put ``x2`` into the batches
            y1_key: Key to put ``y1`` into the batches
            y2_key: Key to put ``y2`` into the batches
            index_key: Key to put samples' ids into the batches

        """
        df = df.copy()

        if extra_data is not None:
            assert all(
                len(record) == len(df) for record in extra_data.values()
            ), "All the extra records need to have the size equal to the dataset's size"

        assert all(x in df.columns for x in (LABELS_COLUMN, PATHS_COLUMN))

        self.input_tensors_key = input_tensors_key
        self.labels_key = labels_key
        self.paths_key = paths_key
        self.categories_key = categories_key if (CATEGORIES_COLUMN in df.columns) else None
        self.sequence_key = sequence_key if (SEQUENCE_COLUMN in df.columns) else None
        self.index_key = index_key

        self.bboxes_exist = all(coord in df.columns for coord in (X1_COLUMN, X2_COLUMN, Y1_COLUMN, Y2_COLUMN))
        if self.bboxes_exist:
            self.x1_key, self.x2_key, self.y1_key, self.y2_key = x1_key, x2_key, y1_key, y2_key
        else:
            self.x1_key, self.x2_key, self.y1_key, self.y2_key = None, None, None, None

        if dataset_root is not None:
            dataset_root = Path(dataset_root)
            df[PATHS_COLUMN] = df[PATHS_COLUMN].apply(lambda x: str(dataset_root / x))
        else:
            df[PATHS_COLUMN] = df[PATHS_COLUMN].astype(str)

        self.df = df
        self.extra_data = extra_data
        self.transform = transform if transform else get_transforms("norm_albu")
        self.f_imread = f_imread or get_im_reader_for_transforms(transform)
        self.read_bytes_image = (
            lru_cache(maxsize=cache_size)(self._read_bytes_image) if cache_size else self._read_bytes_image
        )

        available_augs_types = (albu.Compose, torchvision.transforms.Compose)
        assert isinstance(self.transform, available_augs_types), f"Type of transforms must be in {available_augs_types}"

    @staticmethod
    def _read_bytes_image(path: Union[Path, str]) -> bytes:
        with open(str(path), "rb") as fin:
            return fin.read()

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        row = self.df.iloc[idx]

        img_bytes = self.read_bytes_image(row[PATHS_COLUMN])  # type: ignore
        img = self.f_imread(img_bytes)

        im_h, im_w = img.shape[:2] if isinstance(img, np.ndarray) else img.size[::-1]

        if (not self.bboxes_exist) or any(
            pd.isna(coord) for coord in [row[X1_COLUMN], row[X2_COLUMN], row[Y1_COLUMN], row[Y2_COLUMN]]
        ):
            x1, y1, x2, y2 = 0, 0, im_w, im_h
        else:
            x1, y1, x2, y2 = int(row[X1_COLUMN]), int(row[Y1_COLUMN]), int(row[X2_COLUMN]), int(row[Y2_COLUMN])

        if isinstance(self.transform, albu.Compose):
            img = img[y1:y2, x1:x2, :]  # todo: since albu may handle bboxes we should move it to augs
            image_tensor = self.transform(image=img)["image"]
        else:
            # torchvision.transforms
            img = img.crop((x1, y1, x2, y2))
            image_tensor = self.transform(img)

        item = {
            self.input_tensors_key: image_tensor,
            self.labels_key: row[LABELS_COLUMN],
            self.paths_key: row[PATHS_COLUMN],
            self.index_key: idx,
        }

        if self.categories_key:
            item[self.categories_key] = row[CATEGORIES_COLUMN]

        if self.sequence_key:
            item[self.sequence_key] = row[SEQUENCE_COLUMN]

        if self.bboxes_exist:
            item.update(
                {
                    self.x1_key: x1,
                    self.y1_key: y1,
                    self.x2_key: x2,
                    self.y2_key: y2,
                }
            )

        if self.extra_data:
            for key, record in self.extra_data.items():
                if key in item:
                    raise ValueError(f"<extra_data> and dataset share the same key: {key}")
                else:
                    item[key] = record[idx]

        return item

    def __len__(self) -> int:
        return len(self.df)

    @property
    def bboxes_keys(self) -> Tuple[str, ...]:
        if self.bboxes_exist:
            return self.x1_key, self.y1_key, self.x2_key, self.y2_key
        else:
            return tuple()


class DatasetWithLabels(BaseDataset, IDatasetWithLabels):
    """
    The main purpose of this class is to be used as a dataset during
    the training stage.

    It has to know how to return its labels, which is required information
    to perform the training with the combinations-based losses.
    Particularly, these labels will be passed to `Sampler` to form the batches and
    batches will be passed to `Miner` to form the combinations (triplets).

    """

    def get_labels(self) -> np.ndarray:
        return np.array(self.df[LABELS_COLUMN].tolist())

    def get_label2category(self) -> Optional[Dict[int, Union[str, int]]]:
        """
        Returns:
            Label to category mapping if there was category information in DataFrame, None otherwise.

        """
        if CATEGORIES_COLUMN in self.df.columns:
            label2category = dict(zip(self.df[LABELS_COLUMN], self.df[CATEGORIES_COLUMN]))
        else:
            label2category = None

        return label2category


class DatasetQueryGallery(BaseDataset, IDatasetQueryGallery):
    """
    The main purpose of this class is to be used as a dataset during
    the validation stage. It has to provide information
    about its `query`/`gallery` split.

    Note, that some datasets used as benchmarks in Metric Learning
    provide the splitting information (for example, ``DeepFashion InShop`` dataset), but some of them
    don't (for example, ``CARS196`` or ``CUB200``).
    The validation idea for the latter is to calculate the embeddings for the whole validation set,
    then for every item find ``top-k`` nearest neighbors and calculate the desired retrieval metric.
    In other words, for the desired query item, the gallery is the rest of the validation dataset.

    Thus, if you want to perform this kind of validation process (`1 vs rest`) you should simply return
    ``is_query == True`` and ``is_gallery == True`` for every item in the dataset as the same time.

    """

    def __init__(
        self,
        df: pd.DataFrame,
        extra_data: Optional[Dict[str, Any]] = None,
        dataset_root: Optional[Union[str, Path]] = None,
        transform: Optional[albu.Compose] = None,
        f_imread: Optional[TImReader] = None,
        cache_size: Optional[int] = 0,
        input_tensors_key: str = INPUT_TENSORS_KEY,
        labels_key: str = LABELS_KEY,
        paths_key: str = PATHS_KEY,
        categories_key: str = CATEGORIES_KEY,
        x1_key: str = X1_KEY,
        x2_key: str = X2_KEY,
        y1_key: str = Y1_KEY,
        y2_key: str = Y2_KEY,
        is_query_key: str = IS_QUERY_KEY,
        is_gallery_key: str = IS_GALLERY_KEY,
    ):
        super(DatasetQueryGallery, self).__init__(
            df=df,
            extra_data=extra_data,
            dataset_root=dataset_root,
            transform=transform,
            f_imread=f_imread,
            cache_size=cache_size,
            input_tensors_key=input_tensors_key,
            labels_key=labels_key,
            paths_key=paths_key,
            categories_key=categories_key,
            x1_key=x1_key,
            x2_key=x2_key,
            y1_key=y1_key,
            y2_key=y2_key,
        )
        assert all(x in df.columns for x in (IS_QUERY_COLUMN, IS_GALLERY_COLUMN))

        self.is_query_key = is_query_key
        self.is_gallery_key = is_gallery_key

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        item = super().__getitem__(idx)
        item[self.is_query_key] = bool(self.df.iloc[idx][IS_QUERY_COLUMN])
        item[self.is_gallery_key] = bool(self.df.iloc[idx][IS_GALLERY_COLUMN])
        return item


def get_retrieval_datasets(
    dataset_root: Path,
    transforms_train: Any,
    transforms_val: Any,
    f_imread_train: Optional[TImReader] = None,
    f_imread_val: Optional[TImReader] = None,
    dataframe_name: str = "df.csv",
    cache_size: Optional[int] = 0,
    verbose: bool = True,
) -> Tuple[DatasetWithLabels, DatasetQueryGallery]:
    df = pd.read_csv(dataset_root / dataframe_name, index_col=False)

    check_retrieval_dataframe_format(df, dataset_root=dataset_root, verbose=verbose)

    # first half will consist of "train" split, second one of "val"
    # so labels in train will be from 0 to N-1 and labels in test will be from N to K
    mapper = {l: i for i, l in enumerate(df.sort_values(by=[SPLIT_COLUMN])[LABELS_COLUMN].unique())}

    # train
    df_train = df[df[SPLIT_COLUMN] == "train"].reset_index(drop=True)
    df_train[LABELS_COLUMN] = df_train[LABELS_COLUMN].map(mapper)

    train_dataset = DatasetWithLabels(
        df=df_train,
        dataset_root=dataset_root,
        transform=transforms_train,
        cache_size=cache_size,
        f_imread=f_imread_train,
    )

    # val (query + gallery)
    df_query_gallery = df[df[SPLIT_COLUMN] == "validation"].reset_index(drop=True)
    valid_dataset = DatasetQueryGallery(
        df=df_query_gallery,
        dataset_root=dataset_root,
        transform=transforms_val,
        cache_size=cache_size,
        f_imread=f_imread_val,
    )

    return train_dataset, valid_dataset


__all__ = ["BaseDataset", "DatasetWithLabels", "DatasetQueryGallery", "get_retrieval_datasets"]



from math import ceil
from typing import Any, Optional

import matplotlib.pyplot as plt
import numpy as np
import pytorch_lightning as pl
from pytorch_lightning import Callback
from pytorch_lightning.loggers import NeptuneLogger, TensorBoardLogger, WandbLogger
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch.utils.data import DataLoader

from oml.const import LOG_IMAGE_FOLDER
from oml.ddp.patching import check_loaders_is_patched, patch_dataloader_to_ddp
from oml.interfaces.metrics import IBasicMetric, IMetricDDP, IMetricVisualisable
from oml.lightning.modules.ddp import ModuleDDP
from oml.utils.images.images import figure_to_nparray
from oml.utils.misc import flatten_dict


class MetricValCallback(Callback):
    """
    This is a wrapper which allows to use IBasicMetric with PyTorch Lightning.

    """

    def __init__(
        self,
        metric: IBasicMetric,
        log_images: bool = False,
        loader_idx: int = 0,
        samples_in_getitem: int = 1,
    ):
        """
        Args:
            metric: Metric
            log_images: Set ``True`` if you want to have visual logging
            loader_idx: Idx of the loader to calculate metric for
            samples_in_getitem: Some of the datasets return several samples when calling ``__getitem__``,
                so we need to handle it for the proper calculation. For most of the cases this value equals to 1,
                but for the dataset which explicitly return triplets, this value must be equal to 3,
                for a dataset of pairs it must be equal to 2.

        """

        self.metric = metric
        self.log_images = log_images
        assert not log_images or (
            isinstance(metric, IMetricVisualisable) and metric.ready_to_visualize()  # type: ignore
        )

        self.loader_idx = loader_idx
        self.samples_in_getitem = samples_in_getitem

        self._expected_samples = 0
        self._collected_samples = 0
        self._ready_to_accumulate = False

    def _calc_expected_samples(self, trainer: pl.Trainer, dataloader_idx: int = 0) -> int:
        loaders = (
            [trainer.val_dataloaders] if isinstance(trainer.val_dataloaders, DataLoader) else trainer.val_dataloaders
        )
        return self.samples_in_getitem * len(loaders[dataloader_idx].dataset)

    def on_validation_batch_start(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> None:
        if dataloader_idx == self.loader_idx:
            if not self._ready_to_accumulate:
                self._expected_samples = self._calc_expected_samples(trainer=trainer, dataloader_idx=dataloader_idx)
                self._collected_samples = 0

                self.metric.setup(num_samples=self._expected_samples)
                self._ready_to_accumulate = True

    def on_validation_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Optional[STEP_OUTPUT],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        if dataloader_idx == self.loader_idx:
            assert self._ready_to_accumulate

            self.metric.update_data(outputs)

            self._collected_samples += len(outputs[list(outputs.keys())[0]])
            if self._collected_samples > self._expected_samples:
                self._raise_computation_error()

            # For non-additive metrics (like f1) we usually accumulate some information during the epoch, then we
            # calculate final score `on_validation_epoch_end`. The problem here is that `on_validation_epoch_end`
            # lightning logger doesn't know dataloader_idx and logs metrics in incorrect way if num_dataloaders > 1.
            # To avoid the problem we calculate metrics `on_validation_batch_end` for the last batch in the loader.
            is_last_expected_batch = self._collected_samples == self._expected_samples
            if is_last_expected_batch:
                self.calc_and_log_metrics(pl_module)

    def _log_images(self, pl_module: pl.LightningModule) -> None:
        if not isinstance(self.metric, IMetricVisualisable):
            return

        for fig, metric_log_str in zip(*self.metric.visualize()):
            log_str = f"{LOG_IMAGE_FOLDER}/{metric_log_str}"
            if isinstance(pl_module.logger, NeptuneLogger):
                from neptune.new.types import File  # this is the optional dependency

                pl_module.logger.experiment[log_str].log(File.as_image(fig))
            elif isinstance(pl_module.logger, WandbLogger):
                fig_img = figure_to_nparray(fig)
                pl_module.logger.log_image(images=[fig_img], key=metric_log_str)
            elif isinstance(pl_module.logger, TensorBoardLogger):
                fig_img = figure_to_nparray(fig)
                pl_module.logger.experiment.add_image(
                    log_str, np.transpose(fig_img, (2, 0, 1)), pl_module.current_epoch
                )
            else:
                raise ValueError(f"Logging with {type(pl_module.logger)} is not supported yet.")

            plt.close(fig=fig)

    def on_validation_epoch_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        self._ready_to_accumulate = False
        if self._collected_samples != self._expected_samples:
            self._raise_computation_error()

    def calc_and_log_metrics(self, pl_module: pl.LightningModule) -> None:
        metrics = self.metric.compute_metrics()
        metrics = flatten_dict(metrics)
        pl_module.log_dict(metrics, rank_zero_only=True, add_dataloader_idx=True)

        if self.log_images:
            self._log_images(pl_module=pl_module)

    def _raise_computation_error(self) -> Exception:
        raise ValueError(
            f"Incorrect calculation for {self.metric.__class__.__name__} metric. "
            f"Inconsistent number of samples, obtained: {self._collected_samples}, "
            f"expected: {self._expected_samples}, "
            f"'samples_in_getitem': {self.samples_in_getitem}.\n"
            f"Make sure that you don't use the 'overfit_batches' parameter in 'pl.Trainer' and "
            f"you set 'drop_last=False'. The idea is that lengths of dataset and dataloader must match."
        )


err_message_loaders_is_not_patched = (
    "\nExperiment is runned in DDP mode, but some of validation dataloaders is not patched. Metric callback will "
    "be incorrect  without patched loaders. Possible problems and solutions:\n"
    f"1) If you use custom module inherited from 'pl.LightningModule', please replace ancestor class with "
    f"our '{ModuleDDP.__name__}', which automaticaly patches your loaders\n"
    f"2) If you implement your own 'train_dataloader' or 'val_dataloader' methods for your module, you can add "
    f"extra line of code for patching loader with '{patch_dataloader_to_ddp.__name__}' function\n"
    f"3) If you call 'trainer.fit(...)' or 'trainer.validate(...)' method with loaders as argument, PytorchLightning "
    f"will ignore loaders from 'train_dataloader' and 'val_dataloader' methods. Please avoid substituting loaders to "
    f"this functions, instead use '{ModuleDDP.__name__}'\n"
    f"4) Check that the flag 'replace_sampler_ddp=False' in the trainer constructor, because we do this "
    f"replacement in '{ModuleDDP.__name__}' constructor\n"
    f"5) Turn off the 'overfit_batches' parameter in 'pl.Trainer'."
)


class MetricValCallbackDDP(MetricValCallback):
    """
    This is an extension to the regular callback that takes into account data reduction and padding
    on the inference for each device in DDP setup

    """

    metric: IMetricDDP

    def __init__(self, metric: IMetricDDP, *args: Any, **kwargs: Any):
        assert isinstance(metric, IMetricDDP), "Metric has to support DDP interface"
        super().__init__(metric, *args, **kwargs)

    def _calc_expected_samples(self, trainer: pl.Trainer, dataloader_idx: int = 0) -> int:
        loaders = (
            [trainer.val_dataloaders] if isinstance(trainer.val_dataloaders, DataLoader) else trainer.val_dataloaders
        )
        len_dataset = len(loaders[dataloader_idx].dataset)
        if trainer.world_size > 1:
            # we use padding in DDP and sequential sampler for validation
            len_dataset = ceil(len_dataset / trainer.world_size)
        return self.samples_in_getitem * len_dataset

    def calc_and_log_metrics(self, pl_module: pl.LightningModule) -> None:
        # TODO: optimize to avoid duplication of metrics on all devices.
        #  Note: if we calculate metric only on main device, we need to log (!!!) metric for all devices,
        #  because they need this metric for checkpointing
        self.metric.sync()
        return super().calc_and_log_metrics(pl_module=pl_module)

    @staticmethod
    def _check_loaders(trainer: "pl.Trainer") -> None:
        if trainer.world_size > 1 and trainer.val_dataloaders is not None:
            if not check_loaders_is_patched(trainer.val_dataloaders):
                raise RuntimeError(err_message_loaders_is_not_patched)

    def on_train_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        self._check_loaders(trainer)

    def on_validation_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        self._check_loaders(trainer)


__all__ = ["MetricValCallback", "MetricValCallbackDDP"]



from abc import ABC, abstractmethod
from typing import Any, Dict

import numpy as np
from torch.utils.data import Dataset

from oml.const import (  # noqa
    INDEX_KEY,
    INPUT_TENSORS_KEY,
    IS_GALLERY_KEY,
    IS_QUERY_KEY,
    LABELS_KEY,
    PAIR_1ST_KEY,
    PAIR_2ND_KEY,
)
from oml.samplers.balance import BalanceSampler  # noqa


class IDatasetWithLabels(Dataset, ABC):
    """
    This is an interface for the datasets which can provide their labels.

    """

    input_tensors_key: str = INPUT_TENSORS_KEY
    labels_key: str = LABELS_KEY
    index_key: str = INDEX_KEY

    def __getitem__(self, item: int) -> Dict[str, Any]:
        """

        Args:
            item: Idx of the sample

        Returns:
             Dictionary with the following keys:

            ``self.input_tensors_key``
            ``self.labels_key``
            ``self.index_key``

        """
        raise NotImplementedError()

    @abstractmethod
    def get_labels(self) -> np.ndarray:
        raise NotImplementedError()


class IDatasetQueryGallery(Dataset, ABC):
    """
    This is an interface for the datasets which can provide the information on how to split
    the validation set into the two parts: query and gallery.

    """

    input_tensors_key: str = INPUT_TENSORS_KEY
    labels_key: str = LABELS_KEY
    is_query_key: str = IS_QUERY_KEY
    is_gallery_key: str = IS_GALLERY_KEY
    index_key: str = INDEX_KEY

    @abstractmethod
    def __getitem__(self, item: int) -> Dict[str, Any]:
        """
        Args:
            item: Idx of the sample

        Returns:
             Dictionary with the following keys:

            ``self.input_tensors_key``
            ``self.labels_key``
            ``self.is_query_key``
            ``self.is_gallery_key``
            ``self.index_key``

        """
        raise NotImplementedError()


class IPairsDataset(Dataset, ABC):
    """
    This is an interface for the datasets which return pair of something.

    """

    pairs_1st_key: str = PAIR_1ST_KEY
    pairs_2nd_key: str = PAIR_2ND_KEY
    index_key: str = INDEX_KEY

    @abstractmethod
    def __getitem__(self, item: int) -> Dict[str, Any]:
        """
        Args:
            item: Idx of the sample

        Returns:
             Dictionary with the following keys:

            ``self.pairs_1st_key``
            ``self.pairs_2nd_key``
            ``self.index_key``

        """
        raise NotImplementedError()


__all__ = ["IDatasetWithLabels", "IDatasetQueryGallery", "IPairsDataset"]



from collections import defaultdict
from typing import Any, Dict, Optional

from oml.const import EMBEDDINGS_KEY
from oml.functional.metrics import calculate_accuracy_on_triplets
from oml.interfaces.metrics import IBasicMetric
from oml.utils.misc_torch import OnlineAvgDict


class AccuracyOnTriplets(IBasicMetric):
    metric_name = "accuracy_on_triplets"

    def __init__(self, embeddings_key: str = EMBEDDINGS_KEY, categories_key: Optional[str] = None):
        """

        Args:
            embeddings_key: key to get embeddings which are tensors with the following structure:
                            [anchor, positive, negative, anchor, positive, negative...]
                             with the shape of [n_triplets * 3, features_dim]
            categories_key: key to get categories, which are integers or strings
        """
        self.embeddings_key = embeddings_key
        self.categories_key = categories_key

        self.avg_online = defaultdict(OnlineAvgDict)  # type: ignore

    def setup(self, *args: Any, **kwargs: Any) -> None:  # type: ignore
        self.avg_online = defaultdict(OnlineAvgDict)  # type: ignore

    def update_data(self, data_dict: Dict[str, Any]) -> None:  # type: ignore
        acc_batch = calculate_accuracy_on_triplets(data_dict[self.embeddings_key], reduce_mean=False)

        self.avg_online[self.overall_categories_key].update({self.metric_name: list(map(float, acc_batch))})

        if self.categories_key is not None:
            for acc, category in zip(acc_batch, data_dict[self.categories_key]):
                self.avg_online[category].update({self.metric_name: float(acc)})

    def compute_metrics(self) -> Dict[str, Any]:  # type: ignore
        return {metric: avg_online.get_dict_with_results() for metric, avg_online in self.avg_online.items()}


__all__ = ["AccuracyOnTriplets"]



from typing import Any, Dict, List

from torch import Tensor


class IDistancesPostprocessor:
    """
    This is a parent class for the classes which apply some postprocessing
    after query-to-gallery distance matrix has been calculated.
    For example, we may want to apply one of re-ranking techniques.

    """

    def process(self, distances: Tensor, queries: Any, galleries: Any) -> Tensor:
        """
        This method takes all the needed variables and returns
        the modified matrix of distances, where some distances are
        replaced with new ones.

        Args:
            distances: Matrix with the shape of ``[Q, G]``
            queries: Queries in the amount of ``Q``
            galleries: Galleries in the amount of ``G``

        Returns:
            An updated distances matrix with the shape of ``[Q, G]``

        """
        raise NotImplementedError()

    def process_by_dict(self, distances: Tensor, data: Dict[str, Any]) -> Tensor:
        """
        This method is the analogue of ``process``, but data is passed as a dictionary,
        so we need to use the corresponding keys, which also have to be obtainable by
        ``needed_keys`` property.

        Args:
            distances: Matrix with the shape of ``[Q, G]``
            data: Dictionary of data

        Returns:
            An updated distances matrix with the shape of ``[Q, G]``

        """
        raise NotImplementedError()

    @property
    def needed_keys(self) -> List[str]:
        """
        Returns: Keys that will be used to process data using ``process_by_dict``

        """
        raise NotImplementedError()


__all__ = ["IDistancesPostprocessor"]
